library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;

entity audio_fft is 
	 
	port
	(
		-- LEDS
      LEDG  		: 	out 		std_logic_vector(7 downto 0);
		LEDR			:  out		std_logic_vector(17 downto 0);
                
		--Microphone Input
	--I2C Audio/Video config interface
	 I2C_SCLK 		: out std_logic;
	 I2C_SDAT 		: inout std_logic;
	-- Audio CODEC
	 AUD_XCK 		: out std_logic;
	 AUD_DACLRCK 	: inout std_logic;
	 AUD_ADCLRCK 	: inout std_logic;
	 AUD_BCLK 		: in std_logic;
	 AUD_ADCDAT 	: in std_logic;
	 AUD_DACDAT 	: out std_logic;
		
		GPIO_0		:	out std_logic_vector(7 DOWNTO 0);

      -- CLOCK port
      CLOCK_50 	:	in  		std_logic;
		CLOCK_27 	:	in  		std_logic;
					 
		-- Switch			
		SW				:	in 		std_logic_vector(17 downto 0);
					 
      -- RESET key
      KEY 			:	in  		std_logic_vector(1 downto 0)
      );
end audio_fft;


-- Library Clause(s) (optional)
-- Use Clause(s) (optional)

architecture structure of audio_fft is

component audio_codec 
	port (
	--Inputs
		clk 		: in std_logic;
		reset 		: in std_logic;
		m_read			: in std_logic;
		m_write			: in std_logic;
		writedata_left	: in std_logic_vector( 23 downto 0);
		writedata_right	: in std_logic_vector(23 downto 0);
		AUD_ADCDAT  : in std_logic;
		  --Bidirectional
		AUD_BCLK	: in std_logic;
		AUD_ADCLRCK : inout std_logic;
		AUD_DACLRCK : inout std_logic;
		
		--Outputs
		read_ready, write_ready :  out std_logic;
		readdata_left, readdata_right : out std_logic_vector(23 downto 0);
		AUD_DACDAT : out std_logic
	);		
end component;

component audio_and_video_config 
port(
	-- Inputs
	clk: in std_logic;
	reset 		: in std_logic;

	-- Bidirectionals
	I2C_SDAT : inout std_logic;
	I2C_SCLK : out std_logic
);
end component;

component clock_generator port(
	-- inputs
	CLOCK_27: in std_logic;
	reset: in std_logic;

	-- outputs
	AUD_XCK : out std_logic
);
end component;

component fft_v10_1 is
    port (
      clk          : in  std_logic;
      reset_n      : in  std_logic;
      inverse      : in  std_logic;
      sink_valid   : in  std_logic;
      sink_sop     : in  std_logic;
      sink_eop     : in  std_logic;
      sink_real    : in  std_logic_vector (24 - 1 downto 0);
      sink_imag    : in  std_logic_vector (24 - 1 downto 0);
      source_ready : in  std_logic;
      sink_ready   : out std_logic;
      sink_error   : in  std_logic_vector(1 downto 0);
      source_error : out std_logic_vector(1 downto 0);
      source_sop   : out std_logic;
      source_eop   : out std_logic;
      source_valid : out std_logic;
      source_exp   : out std_logic_vector (5 downto 0);
      source_real  : out std_logic_vector (24 - 1 downto 0);
      source_imag  : out std_logic_vector (24 - 1 downto 0)
      ); 
		
  end component fft_v10_1;
		  
			signal clk,reset_n,read_ready,write_ready,m_read,m_write, rst, inverse,sink_valid,sink_sop,sink_eop,source_ready,sink_ready,source_sop,source_eop,source_valid:std_logic;
		  signal readdata_left, readdata_right, writedata_left,writedata_right,sink_real,sink_imag,source_real,source_imag : std_logic_vector(23 downto 0);
		  signal sink_error,source_error : std_logic_vector(1 downto 0);
		  signal source_exp   : std_logic_vector (5 downto 0);
		  
		  signal clk_cnt,cnt,out_cnt     : natural range 0 to 1000000000;	
		  
		  signal result48,sq_real,sq_imag,max	: signed(47 downto 0);
		  
		  signal temp_real24,temp_imag24,exp_24  : signed(23 downto 0);
		  
		  signal temp_exp : integer;
			
begin

		--the audio codec requires a low enabled reset
		rst <= NOT KEY(1);
		--the fft requires a high enabled reset
		reset_n <= KEY(1);
		  
		--we want a forward fft
		inverse <= '0';
		--running at our generated clock frequency, new data will always be available
		source_ready <= '1';
		sink_valid <= '1';
		--there is no imaginary part to our input data
		sink_imag <= "000000000000000000000000";
		--no errors on the input
		sink_error <= "00";
		--input data is left channel of audio input
		sink_real <= readdata_left;
		--sets up the codec to read as soon as data becomes available
		m_read <= read_ready;
		
		--forwards the audio input to the audio output
		writedata_left <= readdata_left;
		writedata_right <= readdata_left;
		m_write <= write_ready;
		
		
		--generates clock for fft (based on 44.1 kHz sampling frequency of codec
		clkgen : process(CLOCK_50)
		begin
		
			if (rising_edge(CLOCK_50)) then
			clk_cnt <= clk_cnt + 1;
			if(clk_cnt = 1135) then
				clk <= '1';
				clk_cnt <= 0;
			elsif(clk_cnt = 567) then
				clk <= '0';
			end if;
				
			end if;
		end process clkgen;
		  
		fft_in_process: process(clk)
		  
		begin
		  
		if (rising_edge(clk) and sink_ready = '1') then
			
				--indicates the start of packet
				if(cnt = 0) then
					sink_sop <= '1';
				else
					sink_sop <= '0';
				end if;
				
				--counts input packets
				cnt <= cnt + 1;
				
				--indicates end of packet
				if(cnt = 1024) then
					sink_eop <= '1';
					cnt <= 0;
				else
					sink_eop <= '0';
				end if;
				
			end if;
		end process;
		
		fft_out_process: process(clk)
		  
		begin
				if (rising_edge(clk) and source_valid = '1') then
				
				--counts the current bin number of the fft output
				out_cnt <= out_cnt + 1;
			
				--resets bin counter on start of packet
				if(source_sop = '1') then
					out_cnt <= 0;
					--max is used for auto gain control, 
					--this resets it at the start of each output cycle
					max <= signed(conv_std_logic_vector(0,48));
				end if;	
				
				--source_exp is given as a negative number, 
				--this subtracts the exp value from the max value (15)
				temp_exp <= 15 + conv_integer(signed(source_exp));
					
				--shifts the output data to the right by the exp (larger exponents are shifted less)
				temp_real24 <= signed(to_stdlogicvector(to_bitvector(source_real) sra temp_exp));
				--squares the real portion
				sq_real <= temp_real24 * temp_real24;
					
				temp_imag24 <= signed(to_stdlogicvector(to_bitvector(source_imag) sra temp_exp));
				--squares the imaginary portion
				sq_imag <= temp_imag24 * temp_imag24;
					
				--finds the vector magnitude (squared) of the fft output
				result48 <= sq_real + sq_imag;

					--finds max value of all bins
				if(result48 > max) then
					max <= result48;
				end if;
				
				--perform analysis here

			end if;
		  end process;

		AUD : audio_codec port map (
		clk => CLOCK_50,
		reset => rst,
		m_read => m_read,
		m_write => m_write,
		writedata_left => writedata_left,
		writedata_right => writedata_right,
		AUD_ADCDAT => AUD_ADCDAT,
		AUD_BCLK=> AUD_BCLK,
		AUD_ADCLRCK => AUD_ADCLRCK,
		AUD_DACLRCK => AUD_DACLRCK,
		read_ready => read_ready, 
		write_ready => write_ready,
		readdata_left => readdata_left,
		readdata_right => readdata_right,
		AUD_DACDAT => AUD_DACDAT
	);
	
	CFG : audio_and_video_config port map(
		-- Inputs
		clk => CLOCK_50,
		reset => rst,

		-- Bidirectionals
		I2C_SDAT => I2C_SDAT,
		I2C_SCLK => I2C_SCLK
	);
	
	GEN: clock_generator port map(
		-- inputs
		CLOCK_27 => CLOCK_27,
		reset => rst,

		-- outputs
		AUD_XCK => AUD_XCK
	);
	
	FFT : fft_v10_1
    port map (
      clk          => clk,
      reset_n      => reset_n,
      inverse      => inverse,
      sink_valid   => sink_valid,
      sink_sop     => sink_sop,
      sink_eop     => sink_eop,
      sink_real    => sink_real,
      sink_imag    => sink_imag,
      sink_error   => sink_error,
      source_error => source_error,
      source_ready => source_ready,
      sink_ready   => sink_ready,
      source_sop   => source_sop,
      source_eop   => source_eop,
      source_valid => source_valid,
      source_exp   => source_exp,
      source_real  => source_real,
      source_imag  => source_imag);
	
end structure;
